@@ -5,11 +5,10 @@ import pkg_resources
 import tempfile
 import unittest
 
+import josepy as jose
 import mock
 import zope.component
 
-from acme import jose
-
 from certbot import configuration
 
 from certbot.tests import util as test_util
@@ -65,7 +64,7 @@ def get_nginx_configurator(
                     in_progress_dir=os.path.join(backups, "IN_PROGRESS"),
                     server="https://acme-server.org:443/new",
                     tls_sni_01_port=5001,
-                    dry_run=False,
+                    http01_port=80
                 ),
                 name="nginx",
                 version=version)
